@@ -4,12 +4,16 @@ import json
 import logging
 import datetime
 import traceback
+from operator import attrgetter
 
-from sen.exceptions import TerminateApplication
+from sen.constants import ISO_DATETIME_PARSE_STRING
+from sen.exceptions import TerminateApplication, NotifyError, NotAvailableAnymore
 
 import docker
-import humanize
 
+from sen.net import NetData
+from sen.util import calculate_cpu_percent, calculate_blkio_bytes, calculate_network_bytes, repeater, \
+    humanize_time
 
 logger = logging.getLogger(__name__)
 
@@ -103,20 +107,17 @@ def operation(fmt_str):
     def wrap(func):
         @functools.wraps(func)
         def wrapper(self, *args, **kwargs):
-            command_took = None
             pretty_message = ""
-            # cached queries do not access backend -- we don't care about that
-            if kwargs.get("cached", False) is True:
-                response = func(self, *args, **kwargs)
-            else:
-                before = datetime.datetime.now()
-                response = func(self, *args, **kwargs)
-                after = datetime.datetime.now()
 
-                # we want milliseconds, not seconds
-                command_took = (after - before).total_seconds() * 1000
-                logger.debug("%s(%s, %s) %s -> [%f ms]", func.__name__, args, kwargs, self,
-                             command_took)
+            before = datetime.datetime.now()
+            response = func(self, *args, **kwargs)
+            after = datetime.datetime.now()
+
+            # we want milliseconds, not seconds
+            command_took = (after - before).total_seconds() * 1000
+            # # this line literally break zsh; wat?
+            # logger.debug("%s(%s, %s) %s -> [%f ms]", func.__name__, args, kwargs, self,
+            #              command_took)
             if fmt_str:
                 pretty_message = fmt_str.format(object_type=getattr(self, "pretty_object_type", ""),
                                                 object_short_name=getattr(self, "short_name", ""))
@@ -177,21 +178,51 @@ class DockerObject:
     def short_id(self):
         if self._short_id is None:
             self.set_id()
-            self._short_id = self._id[:12]
+            if ":" in self._id:
+                colon_index = self._id.index(":") + 1
+                self._short_id = self._id[colon_index:][:12]
+            else:
+                self._short_id = self._id[:12]
         return self._short_id
 
     def display_time_created(self):
-        return humanize.naturaltime(self.created)
+        return humanize_time(self.created)
 
     def display_formal_time_created(self):
         # http://tools.ietf.org/html/rfc2822.html#section-3.3
         return self.created.strftime("%d %b %Y, %H:%M:%S")
 
-    def inspect(self):
+    def inspect(self, cached=True):
         raise NotImplementedError()
 
     def display_inspect(self):
-        return json.dumps(self.inspect().response, indent=2)
+        try:
+            return json.dumps(self.inspect().response, indent=2)
+        except docker.errors.NotFound:
+            raise NotAvailableAnymore()
+
+    @property
+    def labels(self):
+        labels = self.data["Labels"]
+        return labels
+
+    @property
+    def natural_sort_value(self):
+        return self.created
+
+    def metadata_get(self, path, cached=True):
+        """
+        get metadata from inspect, specified by path
+
+        :param path: list of str
+        :param cached: bool, use cached version of inspect if available
+        """
+        try:
+            value = graceful_chain_get(self.inspect(cached=cached).response, *path)
+        except docker.errors.NotFound:
+            logger.warning("object %s is not available anymore", self)
+            raise NotAvailableAnymore()
+        return value
 
     def __eq__(self, other):
         return type(self) == type(other) and self._id == other._id
@@ -227,8 +258,7 @@ class DockerImage(DockerObject):
         if self.data:
             return self.data.get("ParentId", None)
         else:
-            inspect_operation = self.inspect(cached=True)
-            return inspect_operation.response.get("Parent", None)
+            return self.metadata_get(["Parent"])
 
     @property
     def pretty_object_type(self):
@@ -247,6 +277,33 @@ class DockerImage(DockerObject):
         else:
             return self.docker_backend.scratch_image
 
+    @property
+    def layers(self):
+        """
+        similar as parent images, except that it uses /history API endpoint
+        :return:
+        """
+        # sample output:
+        # {
+        #     "Created": 1457116802,
+        #     "Id": "sha256:507cb13a216097710f0d234668bf64a4c92949c573ba15eba13d05aad392fe04",
+        #     "Size": 204692029,
+        #     "Tags": [
+        #         "docker.io/fedora:latest"
+        #     ],
+        #     "Comment": "",
+        #     "CreatedBy": "/bin/sh -c #(nop) ADD file:bcb5e5c... in /"
+        # }
+        response = self.d.history(self.image_id)
+        layers = []
+        for l in response:
+            layer_id = l["Id"]
+            if layer_id == "<missing>":
+                layers.append(DockerImage(l, self.docker_backend))
+            else:
+                layers.append(self.docker_backend.get_image_by_id(layer_id))
+        return layers
+
     @property
     def children(self):
         return self.docker_backend.get_images_for_parent(self)
@@ -274,19 +331,30 @@ class DockerImage(DockerObject):
 
     @property
     def command(self):
-        cmd = graceful_chain_get(self.inspect(cached=True).response, "Config", "Cmd")
+        cmd = self.metadata_get(["Config", "Cmd"])
         if cmd:
             return " ".join(cmd)
         return ""
 
     @property
     def container_command(self):
-        inspect = self.inspect(cached=True).response
-        cmd = graceful_chain_get(inspect, "ContainerConfig", "Cmd")
-        if cmd:
-            return " ".join(cmd)
+        # history item
+        created_by = graceful_chain_get(self.data, "CreatedBy")
+        if created_by:
+            return created_by
+        try:
+            cmd = self.metadata_get(["ContainerConfig", "Cmd"])
+        except NotAvailableAnymore:
+            pass
+        else:
+            if cmd:
+                return " ".join(cmd)
         return ""
 
+    @property
+    def comment(self):
+        return self.metadata_get(["Comment"])
+
     @property
     def size(self):
         """
@@ -294,12 +362,7 @@ class DockerImage(DockerObject):
 
         :return: int
         """
-        return self.data["VirtualSize"]
-
-    @property
-    def labels(self):
-        labels = self.data["Labels"]
-        return labels
+        return self.data.get("VirtualSize", 0)
 
     @property
     def names(self):
@@ -307,7 +370,8 @@ class DockerImage(DockerObject):
             self._names = []
             if self.data is None:
                 return self._names
-            for t in self.data["RepoTags"]:
+            # RepoTags = image, Tags = output from `history` command, Tags can be None
+            for t in self.data.get("RepoTags", self.data.get("Tags")) or []:
                 image_name = ImageNameStruct.parse(t)
                 if image_name.to_str():
                     self._names.append(image_name)
@@ -354,20 +418,18 @@ class DockerImage(DockerObject):
     def remove(self):
         return self.d.remove_image(self.image_id)
 
-    @operation("tag of {object_type} {object_short_name} removed!")
+    @operation("Tag of {object_type} {object_short_name} removed!")
     def remove_tag(self, tag):
         assert tag in self.names
         return self.d.remove_image(str(tag))
 
     def matches_search(self, s):
-        return s in self.image_id or \
-            s in self.short_name
+        return s in self.image_id or any([s in str(x) for x in self.names])
 
     def __str__(self):
-        if self.names:
-            return "{} ({}) {}".format(self.short_id, ", ".join([x.to_str() for x in self.names]), self.container_command)
-        else:
-            return "{} {}".format(self.short_id, self.container_command)
+        # it's dangerous to put many stuff here b/c most of the values are loaded dynamically
+        # and it's trivial to go into nested exception madness
+        return "image {}".format(self.short_id)
 
     def __repr__(self):
         return self.__str__()
@@ -426,7 +488,7 @@ class DockerContainer(DockerObject):
     def names(self):
         if self._names is None:
             self._names = []
-            for t in self.data["Names"] or []:
+            for t in self.data.get("Names", []):
                 self._names.append(t)
             # sort by name length
             self._names.sort(key=lambda x: len(x))
@@ -444,6 +506,14 @@ class DockerContainer(DockerObject):
     def running(self):
         return self.status.startswith("Up")
 
+    @property
+    def status_created(self):
+        return self.status.startswith("Created")
+
+    @property
+    def exited_well(self):
+        return self.status.startswith("Exited (0)")
+
     @property
     def short_name(self):
         try:
@@ -463,15 +533,69 @@ class DockerContainer(DockerObject):
             image_id = self.data["ImageID"]
         except KeyError:
             # docker <= 1.8
-            image_id = self.inspect(cached=True).response["Image"]
+            image_id = self.metadata_get(["Image"])
         return image_id
 
+    @property
+    def image(self):
+        return self.docker_backend.get_image_by_id(self.image_id)
+
+    @property
+    def ip_address(self):
+        # docker == 1.10
+        ip_address = self.metadata_get(["NetworkSettings", "IPAddress"])
+        return ip_address
+
+    @property
+    def net(self):
+        """
+        get ACTIVE port mappings of a container
+
+        :return: dict:
+        {
+            "host_port": "container_port"
+        }
+        """
+        try:
+            return NetData(self.inspect(cached=True).response)
+        except docker.errors.NotFound:
+            raise NotAvailableAnymore()
+
+    @property
+    def started_at(self):
+        s = self.metadata_get(["State", "StartedAt"])
+        if s:
+            # python expects 6 digits in milliseconds, docker returns 9
+            s = s[:26]
+            if s == "0001-01-01T00:00:00Z":
+                return datetime.datetime.fromordinal(1)
+            started_at = datetime.datetime.strptime(s, ISO_DATETIME_PARSE_STRING)
+            logger.debug("started at %s", started_at)
+            return started_at
+
+    @property
+    def finished_at(self):
+        f = self.metadata_get(["State", "FinishedAt"])
+        if f:
+            f = f[:26]
+            if f == "0001-01-01T00:00:00Z":
+                return datetime.datetime.fromordinal(1)
+            finished_at = datetime.datetime.strptime(f, ISO_DATETIME_PARSE_STRING)
+            logger.debug("finished at %s", finished_at)
+            return finished_at
+
+    @property
+    def natural_sort_value(self):
+        # docker acts weird: 'created' is provided as ordinal and is local time
+        # 'started' and 'finished' are UTC as timestamp; it would be awesome to unite those b/c
+        # atm these inconsistencies mess ordering
+        return max(self.started_at, self.finished_at, super().natural_sort_value)
+
     # methods
 
     def image_name(self):
-        image = self.docker_backend.get_image_by_id(self.image_id)
-        if image is not None:
-            return image.short_name
+        if self.image is not None:
+            return self.image.short_name
         else:
             return self.image_id[:12]
 
@@ -480,6 +604,44 @@ class DockerContainer(DockerObject):
                s in self.short_name
     # api calls
 
+    @operation("Get resources statistics.")
+    def stats(self):
+        for x in self.d.stats(self.container_id, decode=True, stream=True):
+            blk_read, blk_write = calculate_blkio_bytes(x)
+            net_r, net_w = calculate_network_bytes(x)
+            mem_current = x["memory_stats"]["usage"]
+            mem_total = x["memory_stats"]["limit"]
+            r = {
+                "cpu_percent": calculate_cpu_percent(x),
+                "mem_current": mem_current,
+                "mem_total": x["memory_stats"]["limit"],
+                "mem_percent": (mem_current / mem_total) * 100.0,
+                "blk_read": blk_read,
+                "blk_write": blk_write,
+                "net_rx": net_r,
+                "net_tx": net_w,
+            }
+            yield r
+
+    @operation("List processes in running container.")
+    def top(self):
+        """
+        list of processes in a running container
+
+        :return: None or list of dicts
+        """
+        if not self.running:
+            return []
+        # let's get resources from .stats()
+        ps_args = "-eo pid,ppid,wchan,args"
+        # returns {"Processes": [values], "Titles": [values]}
+        # it's easier to play with list of dicts: [{"pid": 1, "ppid": 0}]
+        response = self.d.top(self.container_id, ps_args=ps_args)
+        # TODO: sort?
+        logger.debug(json.dumps(response, indent=2))
+        return [dict(zip(response["Titles"], process))
+                for process in response["Processes"] or []]
+
     @operation("Inspect container {object_short_name}.")
     def inspect(self, cached=False):
         if cached is False or self._inspect is None:
@@ -487,9 +649,9 @@ class DockerContainer(DockerObject):
         return self._inspect
 
     @operation("Logs of container {object_short_name} received.")
-    def logs(self, follow=False):
+    def logs(self, follow=False, lines="all"):
         # when tail is set to all, it takes ages to populate widget
-        logs_data = self.d.logs(self.container_id, stream=follow, tail=16)
+        logs_data = self.d.logs(self.container_id, stream=follow, tail=lines)
         return logs_data
 
     @operation("{object_type} {object_short_name} removed!")
@@ -531,6 +693,7 @@ class DockerBackend:
         self._images = None  # displayed images
         self._all_images = None  # docker images -a
         kwargs = docker.utils.kwargs_from_env(assert_hostname=False)
+        # kwargs["timeout"] = 1  # when debugging timeouts
         try:
             self.client = docker.AutoVersionClient(**kwargs)
         except docker.errors.DockerException as ex:
@@ -544,11 +707,13 @@ class DockerBackend:
         if cached is False or self._images is None:
             logger.debug("doing images() query")
             self._images = {}
-            for i in self.client.images():
+            images_response = repeater(self.client.images) or []
+            for i in images_response:
                 img = DockerImage(i, self)
                 self._images[img.image_id] = img
             self._all_images = {}
-            for i in self.client.images(all=True):
+            all_images_response = repeater(self.client.images, kwargs={"all": True}) or []
+            for i in all_images_response:
                 img = DockerImage(i, self)
                 self._all_images[img.image_id] = img
         return list(self._images.values())
@@ -558,7 +723,8 @@ class DockerBackend:
         if cached is False or self._containers is None:
             logger.debug("doing containers() query")
             self._containers = {}
-            for c in self.client.containers(all=stopped):
+            containers_reponse = repeater(self.client.containers, kwargs={"all": stopped}) or []
+            for c in containers_reponse:
                 container = DockerContainer(c, self)
                 self._containers[container.container_id] = container
         if not stopped:
@@ -566,11 +732,26 @@ class DockerBackend:
         return list(self._containers.values())
 
     def realtime_updates(self):
-        for event in self.client.events(decode=True):
+        it = repeater(self.client.events, kwargs={"decode": True}, retries=5)
+        while True:
+            event = repeater(next, args=(it, ), retries=2)  # likely an engine restart
+            if not event:
+                it = repeater(self.client.events, kwargs={"decode": True}, retries=5)
+                if not it:
+                    raise NotifyError("Unable to fetch realtime updates from docker engine.")
+                continue
             logger.debug("RT event: %s", event)
+            yield event
+            continue
+            # FIXME: this needs rewrite
 
-            # event["from'] means it's a container
-            if "from" in event:
+            try:
+                # 1.10+
+                is_container = event["Type"] == "container"
+            except KeyError:
+                # event["from'] means it's a container
+                is_container = "from" in event
+            if is_container:
                 # inspect doesn't contain info about status and you can't query just one
                 # container with containers()
                 # let's do full-blown containers() query; it's not that expensive
@@ -618,5 +799,5 @@ class DockerBackend:
             images_o = self.get_images(cached=cached)
             content += images_o.response
         if sort_by_created:
-            content.sort(key=lambda x: x.created, reverse=True)
+            content.sort(key=attrgetter("natural_sort_value"), reverse=True)
         return content, containers_o, images_o
